# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal

import httpx

from ..types import (
    ModelID,
    CheckpointsListResponse,
    weight_load_params,
    weight_save_params,
    weight_save_for_sampler_params,
)
from .._types import NOT_GIVEN, Body, Query, Headers, NotGiven, NoneType
from .._utils import maybe_transform, async_maybe_transform
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from .._base_client import make_request_options
from ..types.shared.untyped_api_future import UntypedAPIFuture

__all__ = ["WeightsResource", "AsyncWeightsResource"]


class WeightsResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> WeightsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://docs.hpc-ai.com/sdk/python#accessing-raw-response-data
        """
        return WeightsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> WeightsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://docs.hpc-ai.com/sdk/python#streaming-responses
        """
        return WeightsResourceWithStreamingResponse(self)

    def load(
        self,
        *,
        model_id: ModelID,
        path: str,
        type: Literal["load_weights"] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> UntypedAPIFuture:
        """
        Loads model weights from disk

        Args:
          path: A hpcai URI for model weights at a specific step

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return self._post(
            "/api/v1/load_weights",
            body=maybe_transform(
                {
                    "model_id": model_id,
                    "path": path,
                    "type": type,
                },
                weight_load_params.WeightLoadParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=UntypedAPIFuture,
        )

    def save(
        self,
        *,
        model_id: ModelID,
        path: str | NotGiven = NOT_GIVEN,
        type: Literal["save_weights"] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> UntypedAPIFuture:
        """
        Saves the current model weights to disk

        Args:
          path: A file/directory name for the weights

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return self._post(
            "/api/v1/save_weights",
            body=maybe_transform(
                {
                    "model_id": model_id,
                    "path": path,
                    "type": type,
                },
                weight_save_params.WeightSaveParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=UntypedAPIFuture,
        )

    def save_for_sampler(
        self,
        *,
        model_id: ModelID,
        path: str | NotGiven = NOT_GIVEN,
        type: Literal["save_weights_for_sampler"] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> UntypedAPIFuture:
        """
        Saves weights in a format compatible with sampling/inference servers

        Args:
          path: A file/directory name for the weights

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return self._post(
            "/api/v1/save_weights_for_sampler",
            body=maybe_transform(
                {
                    "model_id": model_id,
                    "path": path,
                    "type": type,
                },
                weight_save_for_sampler_params.WeightSaveForSamplerParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=UntypedAPIFuture,
        )

    def list(
        self,
        model_id: ModelID,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> CheckpointsListResponse:
        """
        Lists available model checkpoints (both training and sampler)

        Args:
          model_id: The model ID to list checkpoints for

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not model_id:
            raise ValueError(f"Expected a non-empty value for `model_id` but received {model_id!r}")
        return self._get(
            f"/api/v1/models/{model_id}/checkpoints",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=CheckpointsListResponse,
        )

    def delete_checkpoint(
        self,
        *,
        model_id: ModelID,
        checkpoint_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> None:
        """Delete a checkpoint for the given training run."""
        if not model_id:
            raise ValueError(f"Expected a non-empty value for `model_id` but received {model_id!r}")
        if not checkpoint_id:
            raise ValueError(
                f"Expected a non-empty value for `checkpoint_id` but received {checkpoint_id!r}"
            )

        self._delete(
            f"/api/v1/training_runs/{model_id}/checkpoints/{checkpoint_id}",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=NoneType,
        )

        return None

class AsyncWeightsResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncWeightsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://docs.hpc-ai.com/sdk/python#accessing-raw-response-data
        """
        return AsyncWeightsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncWeightsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://docs.hpc-ai.com/sdk/python#streaming-responses
        """
        return AsyncWeightsResourceWithStreamingResponse(self)

    async def load(
        self,
        *,
        model_id: ModelID,
        path: str,
        type: Literal["load_weights"] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> UntypedAPIFuture:
        """
        Loads model weights from disk

        Args:
          path: A hpcai URI for model weights at a specific step

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return await self._post(
            "/api/v1/load_weights",
            body=await async_maybe_transform(
                {
                    "model_id": model_id,
                    "path": path,
                    "type": type,
                },
                weight_load_params.WeightLoadParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=UntypedAPIFuture,
        )

    async def save(
        self,
        *,
        model_id: ModelID,
        path: str | NotGiven = NOT_GIVEN,
        type: Literal["save_weights"] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> UntypedAPIFuture:
        """
        Saves the current model weights to disk

        Args:
          path: A file/directory name for the weights

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return await self._post(
            "/api/v1/save_weights",
            body=await async_maybe_transform(
                {
                    "model_id": model_id,
                    "path": path,
                    "type": type,
                },
                weight_save_params.WeightSaveParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=UntypedAPIFuture,
        )

    async def save_for_sampler(
        self,
        *,
        model_id: ModelID,
        path: str | NotGiven = NOT_GIVEN,
        type: Literal["save_weights_for_sampler"] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> UntypedAPIFuture:
        """
        Saves weights in a format compatible with sampling/inference servers

        Args:
          path: A file/directory name for the weights

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return await self._post(
            "/api/v1/save_weights_for_sampler",
            body=await async_maybe_transform(
                {
                    "model_id": model_id,
                    "path": path,
                    "type": type,
                },
                weight_save_for_sampler_params.WeightSaveForSamplerParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=UntypedAPIFuture,
        )

    async def list(
        self,
        model_id: ModelID,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> CheckpointsListResponse:
        """
        Lists available model checkpoints (both training and sampler)

        Args:
          model_id: The model ID to list checkpoints for

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not model_id:
            raise ValueError(f"Expected a non-empty value for `model_id` but received {model_id!r}")
        return await self._get(
            f"/api/v1/training_runs/{model_id}/checkpoints",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=CheckpointsListResponse,
        )

    async def delete_checkpoint(
        self,
        *,
        model_id: ModelID,
        checkpoint_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> None:
        """Delete a checkpoint for the given training run."""
        if not model_id:
            raise ValueError(f"Expected a non-empty value for `model_id` but received {model_id!r}")
        if not checkpoint_id:
            raise ValueError(
                f"Expected a non-empty value for `checkpoint_id` but received {checkpoint_id!r}"
            )

        await self._delete(
            f"/api/v1/training_runs/{model_id}/checkpoints/{checkpoint_id}",
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=NoneType,
        )

        return None


class WeightsResourceWithRawResponse:
    def __init__(self, weights: WeightsResource) -> None:
        self._weights = weights

        self.load = to_raw_response_wrapper(
            weights.load,
        )
        self.save = to_raw_response_wrapper(
            weights.save,
        )
        self.save_for_sampler = to_raw_response_wrapper(
            weights.save_for_sampler,
        )
        self.list = to_raw_response_wrapper(
            weights.list,
        )
        self.delete_checkpoint = to_raw_response_wrapper(
            weights.delete_checkpoint,
        )


class AsyncWeightsResourceWithRawResponse:
    def __init__(self, weights: AsyncWeightsResource) -> None:
        self._weights = weights

        self.load = async_to_raw_response_wrapper(
            weights.load,
        )
        self.save = async_to_raw_response_wrapper(
            weights.save,
        )
        self.save_for_sampler = async_to_raw_response_wrapper(
            weights.save_for_sampler,
        )
        self.list = async_to_raw_response_wrapper(
            weights.list,
        )
        self.delete_checkpoint = async_to_raw_response_wrapper(
            weights.delete_checkpoint,
        )


class WeightsResourceWithStreamingResponse:
    def __init__(self, weights: WeightsResource) -> None:
        self._weights = weights

        self.load = to_streamed_response_wrapper(
            weights.load,
        )
        self.save = to_streamed_response_wrapper(
            weights.save,
        )
        self.save_for_sampler = to_streamed_response_wrapper(
            weights.save_for_sampler,
        )
        self.list = to_streamed_response_wrapper(
            weights.list,
        )
        self.delete_checkpoint = to_streamed_response_wrapper(
            weights.delete_checkpoint,
        )


class AsyncWeightsResourceWithStreamingResponse:
    def __init__(self, weights: AsyncWeightsResource) -> None:
        self._weights = weights

        self.load = async_to_streamed_response_wrapper(
            weights.load,
        )
        self.save = async_to_streamed_response_wrapper(
            weights.save,
        )
        self.save_for_sampler = async_to_streamed_response_wrapper(
            weights.save_for_sampler,
        )
        self.list = async_to_streamed_response_wrapper(
            weights.list,
        )
        self.delete_checkpoint = async_to_streamed_response_wrapper(
            weights.delete_checkpoint,
        )



import logging

import hpcai
import torch

logger = logging.getLogger(__name__)


def compute_mean_nll(
    logprobs_list: list[hpcai.TensorData], weights_list: list[hpcai.TensorData]
) -> float:
    """Compute weighted mean negative log likelihood."""
    total_weighted_logprobs = 0.0
    total_weights = 0.0

    for logprobs, weights in zip(logprobs_list, weights_list, strict=True):
        logprobs_torch = logprobs.to_torch()
        weights_torch = weights.to_torch()
        total_weighted_logprobs += logprobs_torch.dot(weights_torch)
        total_weights += weights_torch.sum()

    if total_weights == 0:
        logger.warning("No valid weights found for NLL computation")
        return float("nan")

    return float(-total_weighted_logprobs / total_weights)


def datum_from_tokens_weights(
    tokens: torch.Tensor,
    weights: torch.Tensor,
    max_length: int | None = None,
) -> hpcai.Datum:
    if max_length is not None:
        tokens = tokens[:max_length]
    weights = weights[:max_length]

    input_tokens = tokens[:-1]
    target_tokens = tokens[1:]
    weights = weights[1:]

    return hpcai.Datum(
        model_input=hpcai.ModelInput.from_ints(tokens=input_tokens.tolist()),
        loss_fn_inputs={
            "weights": hpcai.TensorData(
                data=weights.tolist(),
                dtype="float32",
                shape=list(weights.shape),
            ),
            "target_tokens": hpcai.TensorData(
                data=[int(x) for x in target_tokens.tolist()],
                dtype="int64",
                shape=list(target_tokens.shape),
            ),
        },
    )



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

from __future__ import annotations

from typing import Any
from typing_extensions import override

from ._proxy import LazyProxy


class ResourcesProxy(LazyProxy[Any]):
    """A proxy for the `hpcai.resources` module.

    This is used so that we can lazily import `hpcai.resources` only when
    needed *and* so that users can just import `hpcai` and reference `hpcai.resources`
    """

    @override
    def __load__(self) -> Any:
        import importlib

        mod = importlib.import_module("hpcai.resources")
        return mod


resources = ResourcesProxy().__as_proxied__()



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Any, cast

import httpx

from ..types import ModelID, RequestID, future_retrieve_params
from .._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from .._utils import maybe_transform, async_maybe_transform
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from .._base_client import make_request_options
from ..types.model_id import ModelID
from ..types.request_id import RequestID
from ..types.future_retrieve_response import FutureRetrieveResponse

__all__ = ["FuturesResource", "AsyncFuturesResource"]


class FuturesResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> FuturesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://docs.hpc-ai.com/sdk/python#accessing-raw-response-data
        """
        return FuturesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> FuturesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://docs.hpc-ai.com/sdk/python#streaming-responses
        """
        return FuturesResourceWithStreamingResponse(self)

    def retrieve(
        self,
        *,
        request_id: RequestID,
        model_id: ModelID | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> FutureRetrieveResponse:
        """
        Retrieves the result of a future by its ID

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return cast(
            FutureRetrieveResponse,
            self._post(
                "/api/v1/retrieve_future",
                body=maybe_transform(
                    {
                        "request_id": request_id,
                        "model_id": model_id,
                    },
                    future_retrieve_params.FutureRetrieveParams,
                ),
                options=make_request_options(
                    extra_headers=extra_headers,
                    extra_query=extra_query,
                    extra_body=extra_body,
                    timeout=timeout,
                    idempotency_key=idempotency_key,
                ),
                cast_to=cast(
                    Any, FutureRetrieveResponse
                ),  # Union types cannot be passed in as arguments in the type system
            ),
        )


class AsyncFuturesResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncFuturesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://docs.hpc-ai.com/sdk/python#accessing-raw-response-data
        """
        return AsyncFuturesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncFuturesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://docs.hpc-ai.com/sdk/python#streaming-responses
        """
        return AsyncFuturesResourceWithStreamingResponse(self)

    async def retrieve(
        self,
        *,
        request_id: RequestID,
        model_id: ModelID | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
        max_retries: int | NotGiven = NOT_GIVEN,
    ) -> FutureRetrieveResponse:
        """
        Retrieves the result of a future by its ID

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        options=make_request_options(
            extra_headers=extra_headers,
            extra_query=extra_query,
            extra_body=extra_body,
            timeout=timeout,
            idempotency_key=idempotency_key,
        )
        if not isinstance(max_retries, NotGiven):
            options["max_retries"] = max_retries

        return cast(
            FutureRetrieveResponse,
            await self._post(
                "/api/v1/retrieve_future",
                body=await async_maybe_transform(
                    {
                        "request_id": request_id,
                        "model_id": model_id,
                    },
                    future_retrieve_params.FutureRetrieveParams,
                ),
                options=options,
                cast_to=cast(
                    Any, FutureRetrieveResponse
                ),  # Union types cannot be passed in as arguments in the type system
            ),
        )


class FuturesResourceWithRawResponse:
    def __init__(self, futures: FuturesResource) -> None:
        self._futures = futures

        self.retrieve = to_raw_response_wrapper(
            futures.retrieve,
        )


class AsyncFuturesResourceWithRawResponse:
    def __init__(self, futures: AsyncFuturesResource) -> None:
        self._futures = futures

        self.retrieve = async_to_raw_response_wrapper(
            futures.retrieve,
        )


class FuturesResourceWithStreamingResponse:
    def __init__(self, futures: FuturesResource) -> None:
        self._futures = futures

        self.retrieve = to_streamed_response_wrapper(
            futures.retrieve,
        )


class AsyncFuturesResourceWithStreamingResponse:
    def __init__(self, futures: AsyncFuturesResource) -> None:
        self._futures = futures

        self.retrieve = async_to_streamed_response_wrapper(
            futures.retrieve,
        )



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import httpx

from .._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from .._base_client import make_request_options
from ..types.health_response import HealthResponse
from ..types.get_server_capabilities_response import GetServerCapabilitiesResponse

__all__ = ["ServiceResource", "AsyncServiceResource"]


class ServiceResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> ServiceResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://docs.hpc-ai.com/sdk/python#accessing-raw-response-data
        """
        return ServiceResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> ServiceResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://docs.hpc-ai.com/sdk/python#streaming-responses
        """
        return ServiceResourceWithStreamingResponse(self)

    def get_server_capabilities(
        self,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> GetServerCapabilitiesResponse:
        """Retrieves information about supported models and server capabilities"""
        return self._get(
            "/api/v1/get_server_capabilities",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=GetServerCapabilitiesResponse,
        )

    def health_check(
        self,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> HealthResponse:
        """Checks if the API server is ready"""
        return self._get(
            "/api/v1/healthz",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=HealthResponse,
        )


class AsyncServiceResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncServiceResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://docs.hpc-ai.com/sdk/python#accessing-raw-response-data
        """
        return AsyncServiceResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncServiceResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://docs.hpc-ai.com/sdk/python#streaming-responses
        """
        return AsyncServiceResourceWithStreamingResponse(self)

    async def get_server_capabilities(
        self,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> GetServerCapabilitiesResponse:
        """Retrieves information about supported models and server capabilities"""
        return await self._get(
            "/api/v1/get_server_capabilities",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=GetServerCapabilitiesResponse,
        )

    async def health_check(
        self,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> HealthResponse:
        """Checks if the API server is ready"""
        return await self._get(
            "/api/v1/healthz",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=HealthResponse,
        )


class ServiceResourceWithRawResponse:
    def __init__(self, service: ServiceResource) -> None:
        self._service = service

        self.get_server_capabilities = to_raw_response_wrapper(
            service.get_server_capabilities,
        )
        self.health_check = to_raw_response_wrapper(
            service.health_check,
        )


class AsyncServiceResourceWithRawResponse:
    def __init__(self, service: AsyncServiceResource) -> None:
        self._service = service

        self.get_server_capabilities = async_to_raw_response_wrapper(
            service.get_server_capabilities,
        )
        self.health_check = async_to_raw_response_wrapper(
            service.health_check,
        )


class ServiceResourceWithStreamingResponse:
    def __init__(self, service: ServiceResource) -> None:
        self._service = service

        self.get_server_capabilities = to_streamed_response_wrapper(
            service.get_server_capabilities,
        )
        self.health_check = to_streamed_response_wrapper(
            service.health_check,
        )


class AsyncServiceResourceWithStreamingResponse:
    def __init__(self, service: AsyncServiceResource) -> None:
        self._service = service

        self.get_server_capabilities = async_to_streamed_response_wrapper(
            service.get_server_capabilities,
        )
        self.health_check = async_to_streamed_response_wrapper(
            service.health_check,
        )



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal, Required, TypedDict

from .model_id import ModelID

__all__ = ["TrainingOptimStepParams", "AdamParams"]


class TrainingOptimStepParams(TypedDict, total=False):
    adam_params: Required[AdamParams]

    model_id: Required[ModelID]

    type: Literal["optim_step"] = "optim_step"


class AdamParams(TypedDict, total=False):
    learning_rate: Required[float]
    """Learning rate for the optimizer"""

    beta1: float
    """Coefficient used for computing running averages of gradient"""

    beta2: float
    """Coefficient used for computing running averages of gradient square"""

    eps: float
    """Term added to the denominator to improve numerical stability"""



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Sequence

from typing_extensions import Literal

from .._models import StrictBase

__all__ = ["EncodedTextChunk"]


class EncodedTextChunk(StrictBase):
    tokens: Sequence[int]
    """Array of token IDs"""

    type: Literal["encoded_text"] = "encoded_text"

    @property
    def length(self) -> int:
        return len(self.tokens)



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal, TypeAlias

__all__ = ["EventType"]

EventType: TypeAlias = Literal[
    "SESSION_START", "SESSION_END", "UNHANDLED_EXCEPTION", "GENERIC_EVENT"
]



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

from .._models import BaseModel
from .cursor import Cursor
from .training_run import TrainingRun

__all__ = ["TrainingRunsResponse"]


class TrainingRunsResponse(BaseModel):
    training_runs: list[TrainingRun]
    """List of training runs"""

    cursor: Cursor
    """Pagination cursor information"""



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable
from typing_extensions import Required, TypedDict

from .telemetry_event_param import TelemetryEventParam

__all__ = ["TelemetrySendParams"]


class TelemetrySendParams(TypedDict, total=False):
    events: Required[Iterable[TelemetryEventParam]]

    platform: Required[str]
    """Host platform name"""

    sdk_version: Required[str]
    """SDK version string"""

    session_id: Required[str]



"""ServiceClient for HPC-AI API."""

# Copyright 2025 Thinking Machines Lab
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

from __future__ import annotations

import asyncio
import logging
import os
import time
from typing import TYPE_CHECKING, Any, cast

import hpcai

from hpcai import types
from hpcai.lib.client_connection_pool_type import ClientConnectionPoolType
from hpcai.lib.public_interfaces.api_future import AwaitableConcurrentFuture
from hpcai.lib.telemetry import Telemetry, capture_exceptions
from hpcai.lib.telemetry_provider import TelemetryProvider

from ..api_future_impl import _APIFuture
from ..internal_client_holder import InternalClientHolder
from ..retry_handler import RetryConfig
from ..sync_only import sync_only

if TYPE_CHECKING:
    from .rest_client import RestClient
    from .sampling_client import SamplingClient
    from .training_client import TrainingClient

# pyright: reportPrivateImportUsage=false

logger = logging.getLogger(__name__)


class ModelCreationCancelled(Exception):
    def __init__(self):
        super().__init__("Model creation cancelled by user")

    def __str__(self):
        return "Model creation cancelled by user"

    def __repr__(self):
        return "ModelCreationCancelled()"


class ServiceClient(TelemetryProvider):
    """The ServiceClient is the main entry point for the HPC-AI API. It provides methods to:
    - Query server capabilities and health status
    - Generate TrainingClient instances for model training workflows
    - Generate SamplingClient instances for text generation and inference
    - Generate RestClient instances for REST API operations like listing weights

    Args:
        **kwargs: advanced options passed to the underlying HTTP client,
                 including API keys, headers, and connection settings.

    Example:
        >>> client = ServiceClient()
            # ^^^ near-instant
        >>> training_client = client.create_lora_training_client(base_model="Qwen/Qwen3-8B")
            # ^^^ takes a moment as we initialize the model and assign resources
        >>> sampling_client = client.create_sampling_client(base_model="Qwen/Qwen3-8B")
            # ^^^ near-instant
        >>> rest_client = client.create_rest_client()
            # ^^^ near-instant
    """

    def __init__(self, **kwargs: Any):
        default_headers = _get_default_headers() | kwargs.pop("default_headers", {})
        # Honor caller-provided strict validation flag while defaulting to True
        strict_validation = kwargs.pop("_strict_response_validation", True)
        self.holder = InternalClientHolder(
            **kwargs,
            default_headers=default_headers,
            _strict_response_validation=strict_validation,
        )

    def _get_server_capabilities_submit(
        self,
    ) -> AwaitableConcurrentFuture[types.GetServerCapabilitiesResponse]:
        async def _get_server_capabilities_async():
            async def _send_request():
                with self.holder.aclient(ClientConnectionPoolType.TRAIN) as client:
                    return await client.service.get_server_capabilities()

            return await self.holder.execute_with_retries(_send_request)

        return self.holder.run_coroutine_threadsafe(_get_server_capabilities_async())

    @sync_only
    @capture_exceptions(fatal=True)
    def get_server_capabilities(self) -> types.GetServerCapabilitiesResponse:
        return self._get_server_capabilities_submit().result()

    @capture_exceptions(fatal=True)
    async def get_server_capabilities_async(self) -> types.GetServerCapabilitiesResponse:
        return await self._get_server_capabilities_submit()

    def _create_model_submit(
        self, base_model: str, lora_config: types.LoraConfig
    ) -> AwaitableConcurrentFuture[types.ModelID]:
        async def _create_model_async():
            start_time = time.time()
            retry_count = 0
            QUEUE_RETRY_INTERVAL = 120

            while True:
                try:
                    with self.holder.aclient(ClientConnectionPoolType.TRAIN) as client:
                        future = await client.models.create(
                            base_model=base_model,
                            lora_config=_to_lora_config_params(lora_config)
                        )

                    create_model_response = await _APIFuture(
                        types.CreateModelResponse,
                        self.holder,
                        future,
                        request_start_time=start_time,
                        request_type="CreateModel",
                    ).result_async()

                    return create_model_response.model_id

                except KeyboardInterrupt:
                    logger.info("Model creation request cancelled by user.")
                    raise

                except hpcai.APIStatusError as e:
                    if e.status_code == 423:
                        retry_count += 1
                        if retry_count == 1:
                            logger.warning(
                                "GPU resources are currently at capacity. "
                                "Your request is in the queue and will be processed once resources become available. "
                                "(Press Ctrl+C to cancel)"
                            )
                        await asyncio.sleep(QUEUE_RETRY_INTERVAL)
                        continue

                    raise

        return self.holder.run_coroutine_threadsafe(_create_model_async())

    @sync_only
    @capture_exceptions(fatal=True)
    def create_lora_training_client(
        self,
        base_model: str,
        rank: int = 32,
        seed: int | None = None,
        train_mlp: bool = True,
        train_attn: bool = True,
        train_unembed: bool = True,
    ) -> TrainingClient:
        assert any([train_mlp, train_attn, train_unembed]), (
            "At least one of train_mlp, train_attn, or train_unembed must be True"
        )
        try:
            model_id = self._create_model_submit(
                base_model,
                types.LoraConfig(
                    rank=rank,
                    seed=seed,
                    train_mlp=train_mlp,
                    train_attn=train_attn,
                    train_unembed=train_unembed,
                ),
            ).result()
        except KeyboardInterrupt:
            raise ModelCreationCancelled() from None
        logger.info(f"Creating TrainingClient for {model_id=}")
        return self.create_training_client(model_id)

    @capture_exceptions(fatal=True)
    async def create_lora_training_client_async(
        self,
        base_model: str,
        rank: int = 32,
        seed: int | None = None,
        train_mlp: bool = True,
        train_attn: bool = True,
        train_unembed: bool = True,
    ) -> TrainingClient:
        assert any([train_mlp, train_attn, train_unembed]), (
            "At least one of train_mlp, train_attn, or train_unembed must be True"
        )
        model_id = await self._create_model_submit(
            base_model,
            types.LoraConfig(
                rank=rank,
                seed=seed,
                train_mlp=train_mlp,
                train_attn=train_attn,
                train_unembed=train_unembed,
            ),
        )
        logger.info(f"Creating TrainingClient for {model_id=}")
        return self.create_training_client(model_id)

    @capture_exceptions(fatal=True)
    def create_training_client(self, model_id: types.ModelID | None = None) -> TrainingClient:
        from .training_client import TrainingClient

        return TrainingClient(self.holder, model_id=model_id)

    @sync_only
    @capture_exceptions(fatal=True)
    def create_training_client_from_state(self, path: str) -> TrainingClient:
        rest_client = self.create_rest_client()
        training_run = rest_client.get_training_run_by_checkpoint_path(path).result()

        training_client = self.create_lora_training_client(
            base_model=training_run.base_model,
            rank=training_run.lora_rank,
        )

        training_client.load_state(path).result()
        return training_client

    @capture_exceptions(fatal=True)
    async def create_training_client_from_state_async(self, path: str) -> TrainingClient:
        rest_client = self.create_rest_client()
        training_run = await rest_client.get_training_run_by_checkpoint_path_async(path)

        # Right now all training runs are LoRa runs.
        assert training_run.is_lora and training_run.lora_rank is not None

        training_client = await self.create_lora_training_client_async(
            base_model=training_run.base_model,
            rank=training_run.lora_rank,
        )

        load_future = await training_client.load_state_async(path)
        await load_future.result_async()
        return training_client

    @capture_exceptions(fatal=True)
    def create_sampling_client(
        self,
        model_path: str | None = None,
        base_model: str | None = None,
        retry_config: RetryConfig | None = None,
    ) -> SamplingClient:
        from .sampling_client import SamplingClient

        if model_path is None and base_model is None:
            raise ValueError("Either model_path or base_model must be provided")
        return SamplingClient(
            self.holder,
            model_path=model_path,
            base_model=base_model,
            retry_config=retry_config,
        )

    @capture_exceptions(fatal=True)
    def create_rest_client(self) -> RestClient:
        """Create a RestClient for REST API operations.

        Returns:
            RestClient: A client for listing weights and other REST operations

        Example:
            >>> rest_client = service_client.create_rest_client()
            >>> weights = rest_client.list_model_weights("my-model-id").result()
        """
        from .rest_client import RestClient

        return RestClient(self.holder)

    def get_telemetry(self) -> Telemetry | None:
        return self.holder.get_telemetry()


def _get_default_headers() -> dict[str, str]:
    headers = {}

    if (api_key := os.environ.get("HPCAI_API_KEY", "")) and "X-API-Key" not in headers:
        headers["X-API-Key"] = api_key

    headers["X-Username"] = os.environ.get("USER", "")

    if (
        client_id := os.environ.get("CLOUDFLARE_ACCESS_CLIENT_ID")
    ) and "CF-Access-Client-Id" not in headers:
        headers["CF-Access-Client-Id"] = client_id
    if (
        client_secret := os.environ.get("CLOUDFLARE_ACCESS_CLIENT_SECRET")
    ) and "CF-Access-Client-Secret" not in headers:
        headers["CF-Access-Client-Secret"] = client_secret
    return headers


def _to_lora_config_params(x: types.LoraConfig) -> types._LoraConfigParam:
    return cast(types._LoraConfigParam, x.model_dump())



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from hpcai import HpcAI, AsyncHpcAI
from tests.utils import assert_matches_type
from hpcai.types import TelemetryResponse
from hpcai._utils import parse_datetime

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestTelemetry:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    def test_method_send(self, client: HpcAI) -> None:
        telemetry = client.telemetry.send(
            events=[
                {
                    "event": "SESSION_START",
                    "event_id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "event_session_index": 0,
                    "severity": "DEBUG",
                    "timestamp": parse_datetime("2019-12-27T18:11:19.117Z"),
                }
            ],
            platform="Linux",
            sdk_version="1.2.3",
            session_id="182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
        )
        assert_matches_type(TelemetryResponse, telemetry, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    def test_raw_response_send(self, client: HpcAI) -> None:
        response = client.telemetry.with_raw_response.send(
            events=[
                {
                    "event": "SESSION_START",
                    "event_id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "event_session_index": 0,
                    "severity": "DEBUG",
                    "timestamp": parse_datetime("2019-12-27T18:11:19.117Z"),
                }
            ],
            platform="Linux",
            sdk_version="1.2.3",
            session_id="182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        telemetry = response.parse()
        assert_matches_type(TelemetryResponse, telemetry, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    def test_streaming_response_send(self, client: HpcAI) -> None:
        with client.telemetry.with_streaming_response.send(
            events=[
                {
                    "event": "SESSION_START",
                    "event_id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "event_session_index": 0,
                    "severity": "DEBUG",
                    "timestamp": parse_datetime("2019-12-27T18:11:19.117Z"),
                }
            ],
            platform="Linux",
            sdk_version="1.2.3",
            session_id="182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            telemetry = response.parse()
            assert_matches_type(TelemetryResponse, telemetry, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncTelemetry:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    async def test_method_send(self, async_client: AsyncHpcAI) -> None:
        telemetry = await async_client.telemetry.send(
            events=[
                {
                    "event": "SESSION_START",
                    "event_id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "event_session_index": 0,
                    "severity": "DEBUG",
                    "timestamp": parse_datetime("2019-12-27T18:11:19.117Z"),
                }
            ],
            platform="Linux",
            sdk_version="1.2.3",
            session_id="182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
        )
        assert_matches_type(TelemetryResponse, telemetry, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    async def test_raw_response_send(self, async_client: AsyncHpcAI) -> None:
        response = await async_client.telemetry.with_raw_response.send(
            events=[
                {
                    "event": "SESSION_START",
                    "event_id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "event_session_index": 0,
                    "severity": "DEBUG",
                    "timestamp": parse_datetime("2019-12-27T18:11:19.117Z"),
                }
            ],
            platform="Linux",
            sdk_version="1.2.3",
            session_id="182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        telemetry = await response.parse()
        assert_matches_type(TelemetryResponse, telemetry, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    async def test_streaming_response_send(self, async_client: AsyncHpcAI) -> None:
        async with async_client.telemetry.with_streaming_response.send(
            events=[
                {
                    "event": "SESSION_START",
                    "event_id": "182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
                    "event_session_index": 0,
                    "severity": "DEBUG",
                    "timestamp": parse_datetime("2019-12-27T18:11:19.117Z"),
                }
            ],
            platform="Linux",
            sdk_version="1.2.3",
            session_id="182bd5e5-6e1a-4fe4-a799-aa6d9a6ab26e",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            telemetry = await response.parse()
            assert_matches_type(TelemetryResponse, telemetry, path=["response"])

        assert cast(Any, response.is_closed) is True



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Iterable
from typing_extensions import Required, TypedDict

from .datum_param import DatumParam
from .loss_fn_type import LossFnType

__all__ = ["ForwardBackwardInputParam"]


class ForwardBackwardInputParam(TypedDict, total=False):
    data: Required[Iterable[DatumParam]]
    """Array of input data for the forward/backward pass"""

    loss_fn: Required[LossFnType]
    """Fully qualified function path for the loss function"""



# Copyright 2025 Thinking Machines Lab
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

"""
Utilities for working with tokenizers. Create new types to avoid needing to import AutoTokenizer and PreTrainedTokenizer.


Avoid importing AutoTokenizer and PreTrainedTokenizer until runtime, because they're slow imports.
"""

from __future__ import annotations

from functools import cache
from typing import TYPE_CHECKING, Any, TypeAlias

if TYPE_CHECKING:
    # this import takes a few seconds, so avoid it on the module import when possible
    from transformers.tokenization_utils import PreTrainedTokenizer

    Tokenizer: TypeAlias = PreTrainedTokenizer
else:
    # make it importable from other files as a type in runtime
    Tokenizer: TypeAlias = Any


@cache
def get_tokenizer(model_name: str) -> Tokenizer:
    from transformers.models.auto.tokenization_auto import AutoTokenizer

    # Avoid gating of Llama 3 models:
    if model_name.startswith("meta-llama/Llama-3"):
        model_name = "baseten/Meta-Llama-3-tokenizer"

    return AutoTokenizer.from_pretrained(model_name, use_fast=True)



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

import pytest

from hpcai.lib.chunked_fwdbwd_helpers import (
    _unique,
    _metrics_reduction,
    REDUCE_MAP,
)
from hpcai.types import ForwardBackwardOutput, LossFnOutput, TensorData


class TestMetricsReductionWithUnique:
    """Test the _metrics_reduction function with unique metrics."""

    def create_forward_backward_output(self, metrics: dict, num_loss_fn_outputs: int = 1) -> ForwardBackwardOutput:
        """Helper to create ForwardBackwardOutput for testing."""
        # LossFnOutput is Dict[str, TensorData], so create sample data
        tensor_data = TensorData(
            data=[0.0],
            dtype="float32",
            shape=[1]
        )
        loss_fn_outputs = [{"loss": tensor_data} for _ in range(num_loss_fn_outputs)]
        return ForwardBackwardOutput(
            loss_fn_output_type="test",
            metrics=metrics,
            loss_fn_outputs=loss_fn_outputs,
        )

    def test_unique_reduction_single_result(self):
        """Test unique reduction with single result."""
        results = [
            self.create_forward_backward_output({"clock_cycle:unique": 12345})
        ]

        reduced = _metrics_reduction(results)

        assert "clock_cycle:unique" in reduced
        assert reduced["clock_cycle:unique"] == 12345

    def test_unique_reduction_multiple_results(self):
        """Test unique reduction with multiple results creates additional keys."""
        results = [
            self.create_forward_backward_output({"clock_cycle:unique": 12345}),
            self.create_forward_backward_output({"clock_cycle:unique": 12346}),
            self.create_forward_backward_output({"clock_cycle:unique": 12347}),
        ]

        reduced = _metrics_reduction(results)

        # First value should be the main key
        assert reduced["clock_cycle:unique"] == 12345
        # Additional values should get suffixed keys
        assert reduced["clock_cycle:unique_1"] == 12346
        assert reduced["clock_cycle:unique_2"] == 12347

    def test_unique_reduction_with_other_metrics(self):
        """Test unique reduction alongside other metric types."""
        results = [
            self.create_forward_backward_output({
                "clock_cycle:unique": 100,
                "loss:mean": 0.5,
                "accuracy:max": 0.8,
            }),
            self.create_forward_backward_output({
                "clock_cycle:unique": 101,
                "loss:mean": 0.6,
                "accuracy:max": 0.9,
            }),
        ]

        reduced = _metrics_reduction(results)

        # Unique metric behavior
        assert reduced["clock_cycle:unique"] == 100
        assert reduced["clock_cycle:unique_1"] == 101

        # Other metrics should work as normal
        assert reduced["loss:mean"] == 0.55
        assert reduced["accuracy:max"] == 0.9

    def test_unique_reduction_empty_results(self):
        """Test unique reduction with empty results."""
        results = []
        reduced = _metrics_reduction(results)
        assert reduced == {}

    def test_unique_reduction_missing_metric(self):
        """Test unique reduction when some results don't have the metric."""
        results = [
            self.create_forward_backward_output({"clock_cycle:unique": 100}),
            self.create_forward_backward_output({"other_metric:mean": 0.5}),  # Missing clock_cycle
        ]

        reduced = _metrics_reduction(results)

        # Neither metric should be present since they're not in all results
        # This matches the actual behavior of _metrics_reduction which requires
        # all results to have a metric for it to be processed
        assert "clock_cycle:unique" not in reduced
        assert "other_metric:mean" not in reduced
        assert len(reduced) == 0   # Should be empty

    def test_unique_reduction_with_float_values(self):
        """Test unique reduction with float values."""
        results = [
            self.create_forward_backward_output({"timestamp:unique": 1234567890.123}),
            self.create_forward_backward_output({"timestamp:unique": 1234567890.456}),
        ]

        reduced = _metrics_reduction(results)

        assert reduced["timestamp:unique"] == 1234567890.123
        assert reduced["timestamp:unique_1"] == 1234567890.456

    def test_invalid_reduction_type_raises_assertion(self):
        """Test that invalid reduction types raise AssertionError."""
        results = [
            self.create_forward_backward_output({"invalid:nonexistent": 100})
        ]

        with pytest.raises(AssertionError, match="Invalid reduction nonexistent"):
            _metrics_reduction(results)



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from .._compat import PYDANTIC_V2, ConfigDict
from .._models import StrictBase
from .model_id import ModelID

__all__ = ["SaveWeightsForSamplerRequest"]


class SaveWeightsForSamplerRequest(StrictBase):
    model_id: ModelID

    path: Optional[str] = None
    """A file/directory name for the weights"""

    type: Optional[Literal["save_weights_for_sampler"]] = None



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

"""Public interfaces for the HPC-AI client library."""

from .api_future import APIFuture, AwaitableConcurrentFuture
from .sampling_client import SamplingClient
from .service_client import ServiceClient
from .training_client import TrainingClient

__all__ = [
    "ServiceClient",
    "TrainingClient",
    "SamplingClient",
    "APIFuture",
    "AwaitableConcurrentFuture",
]



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from typing_extensions import TypeAlias

from .encoded_text_chunk_param import EncodedTextChunkParam
from .image_asset_pointer_chunk_param import ImageAssetPointerChunkParam

__all__ = ["ModelInputChunkParam"]

ModelInputChunkParam: TypeAlias = Union[EncodedTextChunkParam, ImageAssetPointerChunkParam]



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

from __future__ import annotations

import inspect
from typing import Any, Callable


def function_has_argument(func: Callable[..., Any], arg_name: str) -> bool:
    """Returns whether or not the given function has a specific parameter"""
    sig = inspect.signature(func)
    return arg_name in sig.parameters


def assert_signatures_in_sync(
    source_func: Callable[..., Any],
    check_func: Callable[..., Any],
    *,
    exclude_params: set[str] = set(),
) -> None:
    """Ensure that the signature of the second function matches the first."""

    check_sig = inspect.signature(check_func)
    source_sig = inspect.signature(source_func)

    errors: list[str] = []

    for name, source_param in source_sig.parameters.items():
        if name in exclude_params:
            continue

        custom_param = check_sig.parameters.get(name)
        if not custom_param:
            errors.append(f"the `{name}` param is missing")
            continue

        if custom_param.annotation != source_param.annotation:
            errors.append(
                f"types for the `{name}` param are do not match; source={repr(source_param.annotation)} checking={repr(custom_param.annotation)}"
            )
            continue

    if errors:
        raise AssertionError(f"{len(errors)} errors encountered when comparing signatures:\n\n" + "\n\n".join(errors))



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Required, TypedDict

from .model_input_param import ModelInputParam
from .loss_fn_inputs_param import LossFnInputsParam

__all__ = ["DatumParam"]


class DatumParam(TypedDict, total=False):
    loss_fn_inputs: Required[LossFnInputsParam]
    """Dictionary mapping field names to tensor data"""

    model_input: Required[ModelInputParam]



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from hpcai import HpcAI, AsyncHpcAI
from tests.utils import assert_matches_type
from hpcai.types.shared import UntypedAPIFuture

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestTraining:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    def test_method_forward(self, client: HpcAI) -> None:
        training = client.training.forward(
            forward_input={
                "data": [
                    {
                        "loss_fn_inputs": {
                            "weights": {
                                "data": [1, 1, 1, 0.5, 0],
                                "dtype": "float32",
                            },
                            "target_tokens": {
                                "data": [123, 456, 789, 101, 202],
                                "dtype": "int64",
                            },
                        },
                        "model_input": {
                            "chunks": [
                                {
                                    "tokens": [1234, 5678, 9012],
                                    "type": "encoded_text",
                                }
                            ]
                        },
                    }
                ],
                "loss_fn": "cross_entropy",
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    def test_method_create(self, client: HpcAI) -> None:
        training = client.training.create(
            base_model="meta-llama/Llama-3.2-1B",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_method_create_with_all_params(self, client: HpcAI) -> None:
        training = client.training.create(
            base_model="meta-llama/Llama-3.2-1B",
            lora_config={"rank": 16},
            type="create_model",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_raw_response_create(self, client: HpcAI) -> None:
        response = client.training.with_raw_response.create(
            base_model="meta-llama/Llama-3.2-1B",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        training = response.parse()
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_streaming_response_create(self, client: HpcAI) -> None:
        with client.training.with_streaming_response.create(
            base_model="meta-llama/Llama-3.2-1B",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            training = response.parse()
            assert_matches_type(UntypedAPIFuture, training, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    def test_method_forward(self, client: HpcAI) -> None:
        pass  # Incomplete test - skipped

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    def test_raw_response_forward(self, client: HpcAI) -> None:
        response = client.training.with_raw_response.forward(
            forward_input={
                "data": [
                    {
                        "loss_fn_inputs": {
                            "weights": {
                                "data": [1, 1, 1, 0.5, 0],
                                "dtype": "float32",
                            },
                            "target_tokens": {
                                "data": [123, 456, 789, 101, 202],
                                "dtype": "int64",
                            },
                        },
                        "model_input": {
                            "chunks": [
                                {
                                    "tokens": [1234, 5678, 9012],
                                    "type": "encoded_text",
                                }
                            ]
                        },
                    }
                ],
                "loss_fn": "cross_entropy",
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        training = response.parse()
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    def test_streaming_response_forward(self, client: HpcAI) -> None:
        with client.training.with_streaming_response.forward(
            forward_input={
                "data": [
                    {
                        "loss_fn_inputs": {
                            "weights": {
                                "data": [1, 1, 1, 0.5, 0],
                                "dtype": "float32",
                            },
                            "target_tokens": {
                                "data": [123, 456, 789, 101, 202],
                                "dtype": "int64",
                            },
                        },
                        "model_input": {
                            "chunks": [
                                {
                                    "tokens": [1234, 5678, 9012],
                                    "type": "encoded_text",
                                }
                            ]
                        },
                    }
                ],
                "loss_fn": "cross_entropy",
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            training = response.parse()
            assert_matches_type(UntypedAPIFuture, training, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    def test_method_forward_backward(self, client: HpcAI) -> None:
        training = client.training.forward_backward(
            forward_backward_input={
                "data": [
                    {
                        "loss_fn_inputs": {
                            "weights": {
                                "data": [1, 1, 1, 0.5, 0],
                                "dtype": "float32",
                            },
                            "target_tokens": {
                                "data": [123, 456, 789, 101, 202],
                                "dtype": "int64",
                            },
                        },
                        "model_input": {
                            "chunks": [
                                {
                                    "tokens": [1234, 5678, 9012],
                                    "type": "encoded_text",
                                }
                            ]
                        },
                    }
                ],
                "loss_fn": "cross_entropy",
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    def test_raw_response_forward_backward(self, client: HpcAI) -> None:
        response = client.training.with_raw_response.forward_backward(
            forward_backward_input={
                "data": [
                    {
                        "loss_fn_inputs": {
                            "weights": {
                                "data": [1, 1, 1, 0.5, 0],
                                "dtype": "float32",
                            },
                            "target_tokens": {
                                "data": [123, 456, 789, 101, 202],
                                "dtype": "int64",
                            },
                        },
                        "model_input": {
                            "chunks": [
                                {
                                    "tokens": [1234, 5678, 9012],
                                    "type": "encoded_text",
                                }
                            ]
                        },
                    }
                ],
                "loss_fn": "cross_entropy",
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        training = response.parse()
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    def test_streaming_response_forward_backward(self, client: HpcAI) -> None:
        with client.training.with_streaming_response.forward_backward(
            forward_backward_input={
                "data": [
                    {
                        "loss_fn_inputs": {
                            "weights": {
                                "data": [1, 1, 1, 0.5, 0],
                                "dtype": "float32",
                            },
                            "target_tokens": {
                                "data": [123, 456, 789, 101, 202],
                                "dtype": "int64",
                            },
                        },
                        "model_input": {
                            "chunks": [
                                {
                                    "tokens": [1234, 5678, 9012],
                                    "type": "encoded_text",
                                }
                            ]
                        },
                    }
                ],
                "loss_fn": "cross_entropy",
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            training = response.parse()
            assert_matches_type(UntypedAPIFuture, training, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    def test_method_optim_step(self, client: HpcAI) -> None:
        training = client.training.optim_step(
            adam_params={"learning_rate": 0},
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    def test_method_optim_step_with_all_params(self, client: HpcAI) -> None:
        training = client.training.optim_step(
            adam_params={
                "learning_rate": 0,
                "beta1": 0,
                "beta2": 0,
                "eps": 0,
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
            type="optim_step",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    def test_raw_response_optim_step(self, client: HpcAI) -> None:
        response = client.training.with_raw_response.optim_step(
            adam_params={"learning_rate": 0},
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        training = response.parse()
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    def test_streaming_response_optim_step(self, client: HpcAI) -> None:
        with client.training.with_streaming_response.optim_step(
            adam_params={"learning_rate": 0},
            model_id="123e4567-e89b-12d3-a456-426614174000",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            training = response.parse()
            assert_matches_type(UntypedAPIFuture, training, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    def test_method_unload(self, client: HpcAI) -> None:
        training = client.training.unload(
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_method_unload_with_all_params(self, client: HpcAI) -> None:
        training = client.training.unload(
            model_id="123e4567-e89b-12d3-a456-426614174000",
            type="unload_model",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_raw_response_unload(self, client: HpcAI) -> None:
        response = client.training.with_raw_response.unload(
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        training = response.parse()
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip()
    @parametrize
    def test_streaming_response_unload(self, client: HpcAI) -> None:
        with client.training.with_streaming_response.unload(
            model_id="123e4567-e89b-12d3-a456-426614174000",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            training = response.parse()
            assert_matches_type(UntypedAPIFuture, training, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncTraining:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    async def test_method_forward(self, async_client: AsyncHpcAI) -> None:
        training = await async_client.training.forward(
            forward_input={
                "data": [
                    {
                        "loss_fn_inputs": {
                            "weights": {
                                "data": [1, 1, 1, 0.5, 0],
                                "dtype": "float32",
                            },
                            "target_tokens": {
                                "data": [123, 456, 789, 101, 202],
                                "dtype": "int64",
                            },
                        },
                        "model_input": {
                            "chunks": [
                                {
                                    "tokens": [1234, 5678, 9012],
                                    "type": "encoded_text",
                                }
                            ]
                        },
                    }
                ],
                "loss_fn": "cross_entropy",
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    async def test_method_create(self, async_client: AsyncHpcAI) -> None:
        training = await async_client.training.create(
            base_model="meta-llama/Llama-3.2-1B",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_method_create_with_all_params(self, async_client: AsyncHpcAI) -> None:
        training = await async_client.training.create(
            base_model="meta-llama/Llama-3.2-1B",
            lora_config={"rank": 16},
            type="create_model",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_raw_response_create(self, async_client: AsyncHpcAI) -> None:
        response = await async_client.training.with_raw_response.create(
            base_model="meta-llama/Llama-3.2-1B",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        training = await response.parse()
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_streaming_response_create(self, async_client: AsyncHpcAI) -> None:
        async with async_client.training.with_streaming_response.create(
            base_model="meta-llama/Llama-3.2-1B",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            training = await response.parse()
            assert_matches_type(UntypedAPIFuture, training, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    async def test_method_forward(self, async_client: AsyncHpcAI) -> None:
        pass  # Incomplete test - skipped

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    async def test_raw_response_forward(self, async_client: AsyncHpcAI) -> None:
        response = await async_client.training.with_raw_response.forward(
            forward_input={
                "data": [
                    {
                        "loss_fn_inputs": {
                            "weights": {
                                "data": [1, 1, 1, 0.5, 0],
                                "dtype": "float32",
                            },
                            "target_tokens": {
                                "data": [123, 456, 789, 101, 202],
                                "dtype": "int64",
                            },
                        },
                        "model_input": {
                            "chunks": [
                                {
                                    "tokens": [1234, 5678, 9012],
                                    "type": "encoded_text",
                                }
                            ]
                        },
                    }
                ],
                "loss_fn": "cross_entropy",
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        training = await response.parse()
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    async def test_streaming_response_forward(self, async_client: AsyncHpcAI) -> None:
        async with async_client.training.with_streaming_response.forward(
            forward_input={
                "data": [
                    {
                        "loss_fn_inputs": {
                            "weights": {
                                "data": [1, 1, 1, 0.5, 0],
                                "dtype": "float32",
                            },
                            "target_tokens": {
                                "data": [123, 456, 789, 101, 202],
                                "dtype": "int64",
                            },
                        },
                        "model_input": {
                            "chunks": [
                                {
                                    "tokens": [1234, 5678, 9012],
                                    "type": "encoded_text",
                                }
                            ]
                        },
                    }
                ],
                "loss_fn": "cross_entropy",
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            training = await response.parse()
            assert_matches_type(UntypedAPIFuture, training, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    async def test_method_forward_backward(self, async_client: AsyncHpcAI) -> None:
        training = await async_client.training.forward_backward(
            forward_backward_input={
                "data": [
                    {
                        "loss_fn_inputs": {
                            "weights": {
                                "data": [1, 1, 1, 0.5, 0],
                                "dtype": "float32",
                            },
                            "target_tokens": {
                                "data": [123, 456, 789, 101, 202],
                                "dtype": "int64",
                            },
                        },
                        "model_input": {
                            "chunks": [
                                {
                                    "tokens": [1234, 5678, 9012],
                                    "type": "encoded_text",
                                }
                            ]
                        },
                    }
                ],
                "loss_fn": "cross_entropy",
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    async def test_raw_response_forward_backward(self, async_client: AsyncHpcAI) -> None:
        response = await async_client.training.with_raw_response.forward_backward(
            forward_backward_input={
                "data": [
                    {
                        "loss_fn_inputs": {
                            "weights": {
                                "data": [1, 1, 1, 0.5, 0],
                                "dtype": "float32",
                            },
                            "target_tokens": {
                                "data": [123, 456, 789, 101, 202],
                                "dtype": "int64",
                            },
                        },
                        "model_input": {
                            "chunks": [
                                {
                                    "tokens": [1234, 5678, 9012],
                                    "type": "encoded_text",
                                }
                            ]
                        },
                    }
                ],
                "loss_fn": "cross_entropy",
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        training = await response.parse()
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    async def test_streaming_response_forward_backward(self, async_client: AsyncHpcAI) -> None:
        async with async_client.training.with_streaming_response.forward_backward(
            forward_backward_input={
                "data": [
                    {
                        "loss_fn_inputs": {
                            "weights": {
                                "data": [1, 1, 1, 0.5, 0],
                                "dtype": "float32",
                            },
                            "target_tokens": {
                                "data": [123, 456, 789, 101, 202],
                                "dtype": "int64",
                            },
                        },
                        "model_input": {
                            "chunks": [
                                {
                                    "tokens": [1234, 5678, 9012],
                                    "type": "encoded_text",
                                }
                            ]
                        },
                    }
                ],
                "loss_fn": "cross_entropy",
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            training = await response.parse()
            assert_matches_type(UntypedAPIFuture, training, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    async def test_method_optim_step(self, async_client: AsyncHpcAI) -> None:
        training = await async_client.training.optim_step(
            adam_params={"learning_rate": 0},
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    async def test_method_optim_step_with_all_params(self, async_client: AsyncHpcAI) -> None:
        training = await async_client.training.optim_step(
            adam_params={
                "learning_rate": 0,
                "beta1": 0,
                "beta2": 0,
                "eps": 0,
            },
            model_id="123e4567-e89b-12d3-a456-426614174000",
            type="optim_step",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    async def test_raw_response_optim_step(self, async_client: AsyncHpcAI) -> None:
        response = await async_client.training.with_raw_response.optim_step(
            adam_params={"learning_rate": 0},
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        training = await response.parse()
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip(reason="Prism tests are disabled")
    @parametrize
    async def test_streaming_response_optim_step(self, async_client: AsyncHpcAI) -> None:
        async with async_client.training.with_streaming_response.optim_step(
            adam_params={"learning_rate": 0},
            model_id="123e4567-e89b-12d3-a456-426614174000",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            training = await response.parse()
            assert_matches_type(UntypedAPIFuture, training, path=["response"])

        assert cast(Any, response.is_closed) is True

    @pytest.mark.skip()
    @parametrize
    async def test_method_unload(self, async_client: AsyncHpcAI) -> None:
        training = await async_client.training.unload(
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_method_unload_with_all_params(self, async_client: AsyncHpcAI) -> None:
        training = await async_client.training.unload(
            model_id="123e4567-e89b-12d3-a456-426614174000",
            type="unload_model",
        )
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_raw_response_unload(self, async_client: AsyncHpcAI) -> None:
        response = await async_client.training.with_raw_response.unload(
            model_id="123e4567-e89b-12d3-a456-426614174000",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        training = await response.parse()
        assert_matches_type(UntypedAPIFuture, training, path=["response"])

    @pytest.mark.skip()
    @parametrize
    async def test_streaming_response_unload(self, async_client: AsyncHpcAI) -> None:
        async with async_client.training.with_streaming_response.unload(
            model_id="123e4567-e89b-12d3-a456-426614174000",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            training = await response.parse()
            assert_matches_type(UntypedAPIFuture, training, path=["response"])

        assert cast(Any, response.is_closed) is True



# Copyright 2025 Thinking Machines Lab
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, Any, Type, Union, Generic, TypeVar, Callable, Optional, cast
from datetime import date, datetime
from typing import TYPE_CHECKING, Any, Callable, Generic, Type, TypeVar, Union, cast

import pydantic
from pydantic.fields import FieldInfo
from typing_extensions import (
    List,
    Unpack,
    Literal,
    ClassVar,
    ParamSpec,
    Protocol,
    Required,
    TypedDict,
    TypeGuard,
    Unpack,
    final,
    runtime_checkable,
)

from ._compat import (
    PYDANTIC_V2,
    ConfigDict,
    field_get_default,
    get_args,
    get_origin,
    is_literal_type,
    is_union,
    parse_obj,
)
from ._compat import (
    GenericModel as BaseGenericModel,
)
from ._constants import RAW_RESPONSE_HEADER
from ._types import (
    AnyMapping,
    Body,
    Headers,
    HttpxRequestFiles,
    NotGiven,
    Query,
    Timeout,
)
from ._utils import (
    PropertyInfo,
    extract_type_arg,
    is_annotated_type,
    is_given,
    is_list,
    is_mapping,
    is_type_alias_type,
    lru_cache,
    parse_date,
    parse_datetime,
    strip_annotated_type,
    strip_not_given,
)

if TYPE_CHECKING:
    from pydantic_core.core_schema import LiteralSchema, ModelField, ModelFieldsSchema, ModelSchema

__all__ = ["StrictBase", "BaseModel", "GenericModel"]

_T = TypeVar("_T")
_BaseModelT = TypeVar("_BaseModelT", bound=pydantic.BaseModel)

P = ParamSpec("P")


@runtime_checkable
class _ConfigProtocol(Protocol):
    allow_population_by_field_name: bool


class StrictBase(pydantic.BaseModel):
    """
    Don't allow extra fields, so user errors are caught earlier.
    Use this for request types.
    """
    model_config = ConfigDict(frozen=True, extra="forbid")

    def __str__(self) -> str:
        return repr(self)

class BaseModel(pydantic.BaseModel):
    """
    Use for classes that may appear in responses. Allow extra fields, so old clients can still work.
    """

    # For future-proofing, we ignore extra fields in case the server adds new fields.
    model_config = ConfigDict(frozen=True, extra="ignore")

    def __str__(self) -> str:
        return repr(self)


def _construct_field(value: object, field: FieldInfo, key: str) -> object:
    if value is None:
        return field_get_default(field)

    if PYDANTIC_V2:
        type_ = field.annotation
    else:
        type_ = cast(type, field.outer_type_)  # type: ignore

    if type_ is None:
        raise RuntimeError(f"Unexpected field type is None for {key}")

    return construct_type(value=value, type_=type_, metadata=getattr(field, "metadata", None))


def is_basemodel(type_: type) -> bool:
    """Returns whether or not the given type is either a `BaseModel` or a union of `BaseModel`"""
    if is_union(type_):
        for variant in get_args(type_):
            if is_basemodel(variant):
                return True

        return False

    return is_basemodel_type(type_)


def is_basemodel_type(type_: type) -> TypeGuard[type[BaseModel] | type[GenericModel]]:
    origin = get_origin(type_) or type_
    if not inspect.isclass(origin):
        return False
    return issubclass(origin, BaseModel) or issubclass(origin, GenericModel)


def build(
    base_model_cls: Callable[P, _BaseModelT],
    *args: P.args,
    **kwargs: P.kwargs,
) -> _BaseModelT:
    """Construct a BaseModel class without validation.

    This is useful for cases where you need to instantiate a `BaseModel`
    from an API response as this provides type-safe params which isn't supported
    by helpers like `construct_type()`.

    ```py
    build(MyModel, my_field_a="foo", my_field_b=123)
    ```
    """
    if args:
        raise TypeError(
            "Received positional arguments which are not supported; Keyword arguments must be used instead",
        )

    return cast(_BaseModelT, construct_type(type_=base_model_cls, value=kwargs))


def construct_type_unchecked(*, value: object, type_: type[_T]) -> _T:
    """Loose coercion to the expected type with construction of nested values.

    Note: the returned value from this function is not guaranteed to match the
    given type.
    """
    return cast(_T, construct_type(value=value, type_=type_))


def construct_type(*, value: object, type_: object, metadata: Optional[List[Any]] = None) -> object:
    """Loose coercion to the expected type with construction of nested values.

    If the given value does not match the expected type then it is returned as-is.
    """

    # store a reference to the original type we were given before we extract any inner
    # types so that we can properly resolve forward references in `TypeAliasType` annotations
    original_type = None

    # we allow `object` as the input type because otherwise, passing things like
    # `Literal['value']` will be reported as a type error by type checkers
    type_ = cast("type[object]", type_)
    if is_type_alias_type(type_):
        original_type = type_  # type: ignore[unreachable]
        type_ = type_.__value__  # type: ignore[unreachable]

    # unwrap `Annotated[T, ...]` -> `T`
    if metadata is not None and len(metadata) > 0:
        meta: tuple[Any, ...] = tuple(metadata)
    elif is_annotated_type(type_):
        meta = get_args(type_)[1:]
        type_ = extract_type_arg(type_, 0)
    else:
        meta = tuple()

    # we need to use the origin class for any types that are subscripted generics
    # e.g. Dict[str, object]
    origin = get_origin(type_) or type_
    args = get_args(type_)

    if is_union(origin):
        try:
            return validate_type(type_=cast("type[object]", original_type or type_), value=value)
        except Exception:
            pass

        # if the type is a discriminated union then we want to construct the right variant
        # in the union, even if the data doesn't match exactly, otherwise we'd break code
        # that relies on the constructed class types, e.g.
        #
        # class FooType:
        #   kind: Literal['foo']
        #   value: str
        #
        # class BarType:
        #   kind: Literal['bar']
        #   value: int
        #
        # without this block, if the data we get is something like `{'kind': 'bar', 'value': 'foo'}` then
        # we'd end up constructing `FooType` when it should be `BarType`.
        discriminator = _build_discriminated_union_meta(union=type_, meta_annotations=meta)
        if discriminator and is_mapping(value):
            variant_value = value.get(discriminator.field_alias_from or discriminator.field_name)
            if variant_value and isinstance(variant_value, str):
                variant_type = discriminator.mapping.get(variant_value)
                if variant_type:
                    return construct_type(type_=variant_type, value=value)

        # if the data is not valid, use the first variant that doesn't fail while deserializing
        for variant in args:
            try:
                return construct_type(value=value, type_=variant)
            except Exception:
                continue

        raise RuntimeError(f"Could not convert data into a valid instance of {type_}")

    if origin == dict:
        if not is_mapping(value):
            return value

        _, items_type = get_args(type_)  # Dict[_, items_type]
        return {key: construct_type(value=item, type_=items_type) for key, item in value.items()}

    if (
        not is_literal_type(type_)
        and inspect.isclass(origin)
        and (issubclass(origin, BaseModel) or issubclass(origin, GenericModel))
    ):
        if is_list(value):
            return [
                construct_type(value=entry, type_=type_) if is_mapping(entry) else entry
                for entry in value
            ]

        if is_mapping(value):
            if issubclass(type_, BaseModel):
                # Convert field types before constructing object
                if PYDANTIC_V2:
                    model_fields = type_.model_fields
                else:
                    model_fields = type_.__fields__  # type: ignore

                converted_values = {}
                for key, val in value.items():
                    if key in model_fields:
                        field = model_fields[key]
                        converted_values[key] = _construct_field(val, field, key)
                    else:
                        converted_values[key] = val

                return type_.construct(**converted_values)  # type: ignore[arg-type]

            return cast(Any, type_).construct(**value)

    if origin == list:
        if not is_list(value):
            return value

        inner_type = args[0]  # List[inner_type]
        return [construct_type(value=entry, type_=inner_type) for entry in value]

    if origin == float:
        if isinstance(value, int):
            coerced = float(value)
            if coerced != value:
                return value
            return coerced

        return value

    if type_ == datetime:
        try:
            return parse_datetime(value)  # type: ignore
        except Exception:
            return value

    if type_ == date:
        try:
            return parse_date(value)  # type: ignore
        except Exception:
            return value

    return value


@runtime_checkable
class CachedDiscriminatorType(Protocol):
    __discriminator__: DiscriminatorDetails


class DiscriminatorDetails:
    field_name: str
    """The name of the discriminator field in the variant class, e.g.

    ```py
    class Foo(BaseModel):
        type: Literal['foo']
    ```

    Will result in field_name='type'
    """

    field_alias_from: str | None
    """The name of the discriminator field in the API response, e.g.

    ```py
    class Foo(BaseModel):
        type: Literal['foo'] = Field(alias='type_from_api')
    ```

    Will result in field_alias_from='type_from_api'
    """

    mapping: dict[str, type]
    """Mapping of discriminator value to variant type, e.g.

    {'foo': FooVariant, 'bar': BarVariant}
    """

    def __init__(
        self,
        *,
        mapping: dict[str, type],
        discriminator_field: str,
        discriminator_alias: str | None,
    ) -> None:
        self.mapping = mapping
        self.field_name = discriminator_field
        self.field_alias_from = discriminator_alias


def _build_discriminated_union_meta(
    *, union: type, meta_annotations: tuple[Any, ...]
) -> DiscriminatorDetails | None:
    if isinstance(union, CachedDiscriminatorType):
        return union.__discriminator__

    discriminator_field_name: str | None = None

    for annotation in meta_annotations:
        if isinstance(annotation, PropertyInfo) and annotation.discriminator is not None:
            discriminator_field_name = annotation.discriminator
            break

    if not discriminator_field_name:
        return None

    mapping: dict[str, type] = {}
    discriminator_alias: str | None = None

    for variant in get_args(union):
        variant = strip_annotated_type(variant)
        if is_basemodel_type(variant):
            if PYDANTIC_V2:
                field = _extract_field_schema_pv2(variant, discriminator_field_name)
                if not field:
                    continue

                # Note: if one variant defines an alias then they all should
                discriminator_alias = field.get("serialization_alias")

                field_schema = field["schema"]

                if field_schema["type"] == "literal":
                    for entry in cast("LiteralSchema", field_schema)["expected"]:
                        if isinstance(entry, str):
                            mapping[entry] = variant
            else:
                field_info = cast("dict[str, FieldInfo]", variant.__fields__).get(
                    discriminator_field_name
                )  # pyright: ignore[reportDeprecated, reportUnnecessaryCast]
                if not field_info:
                    continue

                # Note: if one variant defines an alias then they all should
                discriminator_alias = field_info.alias

                if (annotation := getattr(field_info, "annotation", None)) and is_literal_type(
                    annotation
                ):
                    for entry in get_args(annotation):
                        if isinstance(entry, str):
                            mapping[entry] = variant

    if not mapping:
        return None

    details = DiscriminatorDetails(
        mapping=mapping,
        discriminator_field=discriminator_field_name,
        discriminator_alias=discriminator_alias,
    )
    cast(CachedDiscriminatorType, union).__discriminator__ = details
    return details


def _extract_field_schema_pv2(model: type[BaseModel], field_name: str) -> ModelField | None:
    schema = model.__pydantic_core_schema__
    if schema["type"] == "definitions":
        schema = schema["schema"]

    if schema["type"] != "model":
        return None

    schema = cast("ModelSchema", schema)
    fields_schema = schema["schema"]
    if fields_schema["type"] != "model-fields":
        return None

    fields_schema = cast("ModelFieldsSchema", fields_schema)
    field = fields_schema["fields"].get(field_name)
    if not field:
        return None

    return cast("ModelField", field)  # pyright: ignore[reportUnnecessaryCast]


def validate_type(*, type_: type[_T], value: object) -> _T:
    """Strict validation that the given value matches the expected type"""
    if inspect.isclass(type_) and issubclass(type_, pydantic.BaseModel):
        return cast(_T, parse_obj(type_, value))

    return cast(_T, _validate_non_model_type(type_=type_, value=value))


def set_pydantic_config(typ: Any, config: pydantic.ConfigDict) -> None:
    """Add a pydantic config for the given type.

    Note: this is a no-op on Pydantic v1.
    """
    setattr(typ, "__pydantic_config__", config)  # noqa: B010


# our use of subclassing here causes weirdness for type checkers,
# so we just pretend that we don't subclass
if TYPE_CHECKING:
    GenericModel = BaseModel
else:

    class GenericModel(BaseGenericModel, BaseModel):
        pass


if PYDANTIC_V2:
    from pydantic import TypeAdapter as _TypeAdapter

    _CachedTypeAdapter = cast("TypeAdapter[object]", lru_cache(maxsize=None)(_TypeAdapter))

    if TYPE_CHECKING:
        from pydantic import TypeAdapter
    else:
        TypeAdapter = _CachedTypeAdapter

    def _validate_non_model_type(*, type_: type[_T], value: object) -> _T:
        return TypeAdapter(type_).validate_python(value)

elif not TYPE_CHECKING:  # TODO: condition is weird

    class RootModel(GenericModel, Generic[_T]):
        """Used as a placeholder to easily convert runtime types to a Pydantic format
        to provide validation.

        For example:
        ```py
        validated = RootModel[int](__root__="5").__root__
        # validated: 5
        ```
        """

        __root__: _T

    def _validate_non_model_type(*, type_: type[_T], value: object) -> _T:
        model = _create_pydantic_model(type_).validate(value)
        return cast(_T, model.__root__)

    def _create_pydantic_model(type_: _T) -> Type[RootModel[_T]]:
        return RootModel[type_]  # type: ignore


class FinalRequestOptionsInput(TypedDict, total=False):
    method: Required[str]
    url: Required[str]
    params: Query
    headers: Headers
    max_retries: int
    timeout: float | Timeout | None
    files: HttpxRequestFiles | None
    idempotency_key: str
    json_data: Body
    extra_json: AnyMapping
    follow_redirects: bool


@final
class FinalRequestOptions(pydantic.BaseModel):
    method: str
    url: str
    params: Query = {}
    headers: Union[Headers, NotGiven] = NotGiven()
    max_retries: Union[int, NotGiven] = NotGiven()
    timeout: Union[float, Timeout, None, NotGiven] = NotGiven()
    files: Union[HttpxRequestFiles, None] = None
    idempotency_key: Union[str, None] = None
    post_parser: Union[Callable[[Any], Any], NotGiven] = NotGiven()
    follow_redirects: Union[bool, None] = None

    # It should be noted that we cannot use `json` here as that would override
    # a BaseModel method in an incompatible fashion.
    json_data: Union[Body, None] = None
    extra_json: Union[AnyMapping, None] = None

    if PYDANTIC_V2:
        model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)
    else:

        class Config(pydantic.BaseConfig):  # pyright: ignore[reportDeprecated]
            arbitrary_types_allowed: bool = True

    def get_max_retries(self, max_retries: int) -> int:
        if isinstance(self.max_retries, NotGiven):
            return max_retries
        return self.max_retries

    def _strip_raw_response_header(self) -> None:
        if not is_given(self.headers):
            return

        if self.headers.get(RAW_RESPONSE_HEADER):
            self.headers = {**self.headers}
            self.headers.pop(RAW_RESPONSE_HEADER)

    # override the `construct` method so that we can run custom transformations.
    # this is necessary as we don't want to do any actual runtime type checking
    # (which means we can't use validators) but we do want to ensure that `NotGiven`
    # values are not present
    #
    # type ignore required because we're adding explicit types to `**values`
    @classmethod
    def construct(  # type: ignore
        cls,
        _fields_set: set[str] | None = None,
        **values: Unpack[FinalRequestOptionsInput],
    ) -> FinalRequestOptions:
        kwargs: dict[str, Any] = {
            # we unconditionally call `strip_not_given` on any value
            # as it will just ignore any non-mapping types
            key: strip_not_given(value)
            for key, value in values.items()
        }
        if PYDANTIC_V2:
            return super().model_construct(_fields_set, **kwargs)
        return cast(FinalRequestOptions, super().construct(_fields_set, **kwargs))  # pyright: ignore[reportDeprecated]

    if not TYPE_CHECKING:
        # type checkers incorrectly complain about this assignment
        model_construct = construct



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from .._compat import PYDANTIC_V2, ConfigDict
from .._models import StrictBase
from .model_id import ModelID

__all__ = ["OptimStepRequest", "AdamParams"]


class AdamParams(StrictBase):
    learning_rate: float = 0.0001
    """Learning rate for the optimizer"""

    beta1: float = 0.9
    """Coefficient used for computing running averages of gradient"""

    beta2: float = 0.95
    """Coefficient used for computing running averages of gradient square"""

    eps: float = 1e-12
    """Term added to the denominator to improve numerical stability"""


class OptimStepRequest(StrictBase):
    adam_params: AdamParams

    model_id: ModelID

    type: Optional[Literal["optim_step"]] = None



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

"""
Mock API server for the HpcAI Python SDK.

This server provides mock implementations of all HpcAI API endpoints for testing purposes.
"""

import random
import uuid
import traceback
import logging
from datetime import datetime
from typing import Any, Dict, List

import numpy as np
from fastapi import FastAPI, HTTPException, Request
from fastapi.responses import JSONResponse
from fastapi.exceptions import RequestValidationError
from pydantic import ValidationError

# Import types from hpcai
from hpcai import types

# Configure logging
logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

app = FastAPI(title="HpcAI Mock API Server")

# Custom exception handler to log stack traces
@app.exception_handler(Exception)
async def log_exceptions(request: Request, exc: Exception):
    """Log all exceptions with full stack traces."""
    logger.error(f"Unhandled exception in {request.method} {request.url}")
    logger.error(f"Exception type: {type(exc).__name__}")
    logger.error(f"Exception message: {str(exc)}")
    logger.error(f"Full traceback:\n{traceback.format_exc()}")

    return JSONResponse(
        status_code=500,
        content={"detail": f"Internal server error: {str(exc)}"}
    )


# Handler for validation errors (422)
@app.exception_handler(RequestValidationError)
async def validation_exception_handler(request: Request, exc: RequestValidationError):
    """Log validation errors with details."""
    logger.error(f"Validation error in {request.method} {request.url}")
    logger.error(f"Validation errors: {exc.errors()}")
    logger.error(f"Request body: {exc.body}")
    logger.error(f"Full traceback:\n{traceback.format_exc()}")

    return JSONResponse(
        status_code=422,
        content={
            "detail": exc.errors(),
            "body": exc.body,
            "debug_info": {
                "url": str(request.url),
                "method": request.method,
                "errors": exc.errors()
            }
        }
    )

# Handler for Pydantic validation errors
@app.exception_handler(ValidationError)
async def pydantic_validation_exception_handler(request: Request, exc: ValidationError):
    """Log Pydantic validation errors with details."""
    logger.error(f"Pydantic validation error in {request.method} {request.url}")
    logger.error(f"Validation errors: {exc.errors()}")
    logger.error(f"Full traceback:\n{traceback.format_exc()}")

    return JSONResponse(
        status_code=422,
        content={
            "detail": exc.errors(),
            "debug_info": {
                "url": str(request.url),
                "method": request.method,
                "errors": exc.errors()
            }
        }
    )
# In-memory storage for futures and their results
futures_store: Dict[str, Any] = {}

# In-memory storage for LoRA adapters
lora_adapters: Dict[str, Dict[str, Any]] = {}

# Mock model configurations
SUPPORTED_MODELS = [
    {"model_id": "llama-3-8b", "model_name": "meta-llama/Meta-Llama-3-8B", "arch": "llama"},
    {"model_id": "llama-3-70b", "model_name": "meta-llama/Meta-Llama-3-70B", "arch": "llama"},
    {"model_id": "qwen2-72b", "model_name": "Qwen/Qwen2-72B", "arch": "qwen2"},
]


def generate_future_id() -> str:
    """Generate a unique future ID."""
    return f"future_{uuid.uuid4().hex[:8]}"


@app.get("/healthz", response_model=types.HealthResponse)
async def health_check():
    """Health check endpoint."""
    return types.HealthResponse(status="ok")


@app.get("/get_server_capabilities", response_model=types.GetServerCapabilitiesResponse)
async def get_server_capabilities():
    """Get server capabilities including supported models."""
    supported_models = [
        {"model_id": model["model_id"], "model_name": model["model_name"], "arch": model["arch"]}
        for model in SUPPORTED_MODELS
    ]
    return types.GetServerCapabilitiesResponse(supported_models=supported_models)

def generate_mock_logprobs(seq_len: int) -> List[float]:
    """Generate mock log probabilities for a sequence."""
    return [random.uniform(-4.0, 0.0) for _ in range(seq_len)]

def generate_mock_loss() -> float:
    """Generate mock loss."""
    return random.uniform(0.5, 3.0)


def chunk_length(chunk: types.ModelInputChunkParam) -> int:
    match chunk["type"]:
        case "encoded_text":
            return len(chunk["tokens"])
        case "image_asset_pointer":
            return chunk["tokens"]
        case _:
            raise ValueError(f"Unknown chunk type: {chunk['type']}")

def sequence_length(model_input: types.ModelInputParam) -> int:
    return sum(chunk_length(chunk) for chunk in model_input["chunks"])


@app.post("/fwd", response_model=types.UntypedAPIFuture)
async def forward(params: types.TrainingForwardParams):
    """Perform forward pass."""
    future_id = generate_future_id()

    result = types.FwdBwdOutput(
        loss_fn_outputs={
            "logprobs": [generate_mock_logprobs(sequence_length(datum.input_sequence)) for datum in params.fwdbwd_input.data]
        },
        metrics={
            "loss": generate_mock_loss(),
            "perplexity": generate_mock_loss(),
        }
    )

    # Store the result for future retrieval
    futures_store[future_id] = {
        "result": result,
        "status": "completed",
        "created_at": datetime.now().isoformat(),
    }

    return types.UntypedAPIFuture(request_id=future_id, model_id=params.get("model_id"))


@app.post("/fwdbwd", response_model=types.UntypedAPIFuture)
async def forward_backward(params: types.TrainingForwardBackwardParams):
    """Perform forward and backward pass."""
    # Since the mock implementation is identical, we can reuse the forward logic
    # In a real implementation, forward_backward would also compute gradients
    return await forward(params)


@app.post("/optim_step", response_model=types.UntypedAPIFuture)
async def optim_step(params: types.TrainingOptimStepParams):
    """Perform optimization step."""
    future_id = generate_future_id()

    # Mock optimization step result (OptimStepResponse is just a Dict[str, Union[float, str]])
    result = {
        "grad_norm": random.uniform(0.1, 10.0),
        "weight_norm": random.uniform(10.0, 100.0),
        "update_norm": random.uniform(0.001, 0.1),
    }

    # Store the result for future retrieval
    futures_store[future_id] = {
        "result": result,
        "status": "completed",
        "created_at": datetime.now().isoformat(),
    }

    return UntypedAPIFuture(request_id=future_id, model_id=params.get("model_id"))

@app.post("/retrieve_future")
async def retrieve_future(params: types.FutureRetrieveParams):
    """Retrieve the result of a future."""
    future_id = params["request_id"]

    if future_id not in futures_store:
        raise HTTPException(status_code=404, detail=f"Future {future_id} not found")

    future_data = futures_store[future_id]
    result = future_data["result"]

    # Handle different result types explicitly
    if isinstance(result, (types.FwdBwdOutput, types.AddLoraResponse, types.UnloadModelResponse,
                          types.LoadWeightsResponse, types.SaveWeightsResponse,
                          types.SaveWeightsForSamplerResponse)):
        serialized_result = result.model_dump()
        print(f"RETRIEVE_FUTURE: Returning Pydantic model result: {serialized_result}")
        return serialized_result
    else:
        # For dict results (like OptimStepResponse)
        print(f"RETRIEVE_FUTURE: Returning dict result: {result}")
        return result

@app.post("/add_lora", response_model=types.UntypedAPIFuture)
async def add_lora(params: types.LoraAddParams):
    """Add a LoRA adapter to the model."""
    future_id = generate_future_id()

    # Generate new model_id with LoRA
    base_model = params["base_model"]
    lora_model_id = f"{base_model}_lora_{uuid.uuid4().hex[:8]}"

    # Store LoRA configuration
    if base_model not in lora_adapters:
        lora_adapters[base_model] = {}

    lora_adapters[base_model][lora_model_id] = {
        "rank": params.get("rank", 16),
        "alpha": params.get("alpha", 32),
        "created_at": datetime.now().isoformat()
    }

    # Create the result that will be retrieved later
    result = types.AddLoraResponse(model_id=lora_model_id)

    # Store the result for future retrieval
    futures_store[future_id] = {
        "result": result,
        "status": "completed",
        "created_at": datetime.now().isoformat()
    }

    return types.UntypedAPIFuture(
        request_id=future_id,
        model_id=lora_model_id
    )


@app.post("/remove_lora", response_model=types.UntypedAPIFuture)
async def remove_lora(params: types.LoraRemoveParams):
    """Remove a LoRA adapter from the model."""
    future_id = generate_future_id()

    model_id = params["model_id"]

    # Check if this is a LoRA model
    assert "_lora_" in model_id, f"Model {model_id} is not a LoRA model"

    # Remove from our tracking
    base_model_id = model_id.split("_lora_")[0]
    if base_model_id in lora_adapters and model_id in lora_adapters[base_model_id]:
        del lora_adapters[base_model_id][model_id]


    # Store the result for future retrieval
    futures_store[future_id] = {
        "result": result,
        "status": "completed",
        "created_at": datetime.now().isoformat()
    }

    return types.UntypedAPIFuture(
        request_id=future_id,
        model_id=params.get("model_id")
    )


@app.post("/load_weights", response_model=types.UntypedAPIFuture)
async def load_weights(params: types.WeightLoadParams):
    """Load model weights from a path."""
    future_id = generate_future_id()

    # Mock implementation - in reality this would load weights from storage
    result = LoadWeightsResponse(message=f"Weights loaded from {params['path']}", success=True)

    # Store the result for future retrieval
    futures_store[future_id] = {
        "result": result,
        "status": "completed",
        "created_at": datetime.now().isoformat(),
    }

    return UntypedAPIFuture(request_id=future_id, model_id=params.get("model_id"))


@app.post("/save_weights", response_model=types.UntypedAPIFuture)
async def save_weights(params: types.WeightSaveParams):
    """Save model weights to a path."""
    future_id = generate_future_id()

    # Mock implementation - in reality this would save weights to storage
    save_path = (
        f"{params.get('path', '/tmp')}/checkpoint_{datetime.now().strftime('%Y%m%d_%H%M%S')}"
    )

    result = SaveWeightsResponse(message=f"Weights saved to {save_path}", success=True)

    # Store the result for future retrieval
    futures_store[future_id] = {
        "result": result,
        "status": "completed",
        "created_at": datetime.now().isoformat(),
    }

    return UntypedAPIFuture(request_id=future_id, model_id=params.get("model_id"))


@app.post("/save_weights_for_sampler", response_model=types.UntypedAPIFuture)
async def save_weights_for_sampler(params: types.WeightSaveForSamplerParams):
    """Save weights in a format suitable for the sampler."""
    future_id = generate_future_id()

    # Mock implementation
    save_path = (
        f"{params.get('path', '/tmp')}/sampler_weights_{datetime.now().strftime('%Y%m%d_%H%M%S')}"
    )

    result = types.SaveWeightsForSamplerResponse(
        message=f"Sampler weights saved to {save_path}", success=True
    )

    # Store the result for future retrieval
    futures_store[future_id] = {
        "result": result,
        "status": "completed",
        "created_at": datetime.now().isoformat(),
    }

    return types.UntypedAPIFuture(request_id=future_id, model_id=params.get("model_id"))


@app.post("/get_info", response_model=types.GetInfoResponse)
async def get_info(params: types.ModelGetInfoParams):
    """Get information about a model."""
    model_id = params["model_id"]

    # Find the model in our supported models or check if it's a LoRA model
    model_info = None
    if "_lora_" in model_id:
        # Extract base model ID from LoRA model
        base_model_id = model_id.split("_lora_")[0]
        model_info = next((m for m in SUPPORTED_MODELS if m["model_id"] == base_model_id), None)
    else:
        model_info = next((m for m in SUPPORTED_MODELS if m["model_id"] == model_id), None)

    if not model_info:
        # Default model info for unknown models
        model_info = {"model_name": f"unknown/{model_id}", "arch": "unknown"}

    return types.GetInfoResponse(
        model_data={"model_name": model_info["model_name"], "arch": model_info["arch"]},
        model_id=model_id,
    )


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, host="0.0.0.0", port=8000)



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional

from .._models import BaseModel
from .stop_reason import StopReason

__all__ = ["SampledSequence"]


class SampledSequence(BaseModel):
    stop_reason: StopReason
    """Reason why sampling stopped"""

    tokens: List[int]
    """List of generated token IDs"""

    logprobs: Optional[List[float]] = None
    """Log probabilities for each token (optional)"""



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal

import httpx

from ..types import sampling_sample_params, sampling_asample_params
from .._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from .._utils import maybe_transform, async_maybe_transform
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from .._base_client import make_request_options
from ..types.sample_response import SampleResponse
from ..types.model_input_param import ModelInputParam
from ..types.sampling_params_param import SamplingParamsParam
from ..types.shared.untyped_api_future import UntypedAPIFuture

__all__ = ["SamplingResource", "AsyncSamplingResource"]


class SamplingResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> SamplingResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://docs.hpc-ai.com/sdk/python#accessing-raw-response-data
        """
        return SamplingResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> SamplingResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://docs.hpc-ai.com/sdk/python#streaming-responses
        """
        return SamplingResourceWithStreamingResponse(self)

    def asample(
        self,
        *,
        num_samples: int = 1,
        prompt: ModelInputParam,
        sampling_params: SamplingParamsParam,
        base_model: str | NotGiven = NOT_GIVEN,
        model_path: str | NotGiven = NOT_GIVEN,
        prompt_logprobs: bool | NotGiven = NOT_GIVEN,
        type: Literal["sample"] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> UntypedAPIFuture:
        """
        Generates samples from the model using the specified sampling parameters

        Args:
          num_samples: Number of samples to generate

          base_model: Optional base model name to sample from. Is inferred from model_path, if
              provided. If sampling against a base model, this is required.

          model_path: Optional hpcai:// path to your model weights or LoRA weights. If not provided,
              samples against the base model.

          prompt_logprobs: If set to `true`, computes and returns logprobs on the prompt tokens. Defaults
              to false.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return self._post(
            "/api/v1/asample",
            body=maybe_transform(
                {
                    "num_samples": num_samples,
                    "prompt": prompt,
                    "sampling_params": sampling_params,
                    "base_model": base_model,
                    "model_path": model_path,
                    "prompt_logprobs": prompt_logprobs,
                    "type": type,
                },
                sampling_asample_params.SamplingAsampleParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=UntypedAPIFuture,
        )

    def sample(
        self,
        *,
        num_samples: int = 1,
        prompt: ModelInputParam,
        sampling_params: SamplingParamsParam,
        base_model: str | NotGiven = NOT_GIVEN,
        model_path: str | NotGiven = NOT_GIVEN,
        prompt_logprobs: bool | NotGiven = NOT_GIVEN,
        type: Literal["sample"] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
    ) -> SampleResponse:
        """
        Generates samples from the model using the specified sampling parameters

        Args:
          num_samples: Number of samples to generate

          base_model: Optional base model name to sample from. Is inferred from model_path, if
              provided. If sampling against a base model, this is required.

          model_path: Optional hpcai:// path to your model weights or LoRA weights. If not provided,
              samples against the base model.

          prompt_logprobs: If set to `true`, computes and returns logprobs on the prompt tokens. Defaults
              to false.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        return self._post(
            "/api/v1/sample",
            body=maybe_transform(
                {
                    "num_samples": num_samples,
                    "prompt": prompt,
                    "sampling_params": sampling_params,
                    "base_model": base_model,
                    "model_path": model_path,
                    "prompt_logprobs": prompt_logprobs,
                    "type": type,
                },
                sampling_sample_params.SamplingSampleParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
            ),
            cast_to=SampleResponse,
        )


class AsyncSamplingResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncSamplingResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://docs.hpc-ai.com/sdk/python#accessing-raw-response-data
        """
        return AsyncSamplingResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncSamplingResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://docs.hpc-ai.com/sdk/python#streaming-responses
        """
        return AsyncSamplingResourceWithStreamingResponse(self)

    async def asample(
        self,
        *,
        num_samples: int = 1,
        prompt: ModelInputParam,
        sampling_params: SamplingParamsParam,
        base_model: str | NotGiven = NOT_GIVEN,
        model_path: str | NotGiven = NOT_GIVEN,
        prompt_logprobs: bool | NotGiven = NOT_GIVEN,
        type: Literal["sample"] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
        max_retries: int | NotGiven = NOT_GIVEN,
    ) -> UntypedAPIFuture:
        """
        Generates samples from the model using the specified sampling parameters

        Args:
          num_samples: Number of samples to generate

          base_model: Optional base model name to sample from. Is inferred from model_path, if
              provided. If sampling against a base model, this is required.

          model_path: Optional hpcai:// path to your model weights or LoRA weights. If not provided,
              samples against the base model.

          prompt_logprobs: If set to `true`, computes and returns logprobs on the prompt tokens. Defaults
              to false.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        options = make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
        )
        if max_retries is not NOT_GIVEN:
            options["max_retries"] = max_retries

        return await self._post(
            "/api/v1/asample",
            body=await async_maybe_transform(
                {
                    "num_samples": num_samples,
                    "prompt": prompt,
                    "sampling_params": sampling_params,
                    "base_model": base_model,
                    "model_path": model_path,
                    "prompt_logprobs": prompt_logprobs,
                    "type": type,
                },
                sampling_asample_params.SamplingAsampleParams,
            ),
            options=options,
            cast_to=UntypedAPIFuture,
        )

    async def sample(
        self,
        *,
        num_samples: int = 1,
        prompt: ModelInputParam,
        sampling_params: SamplingParamsParam,
        base_model: str | NotGiven = NOT_GIVEN,
        model_path: str | NotGiven = NOT_GIVEN,
        prompt_logprobs: bool | NotGiven = NOT_GIVEN,
        type: Literal["sample"] | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
        idempotency_key: str | None = None,
        max_retries: int | NotGiven = NOT_GIVEN,
    ) -> SampleResponse:
        """
        Generates samples from the model using the specified sampling parameters

        Args:
          num_samples: Number of samples to generate

          base_model: Optional base model name to sample from. Is inferred from model_path, if
              provided. If sampling against a base model, this is required.

          model_path: Optional hpcai:// path to your model weights or LoRA weights. If not provided,
              samples against the base model.

          prompt_logprobs: If set to `true`, computes and returns logprobs on the prompt tokens. Defaults
              to false.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds

          idempotency_key: Specify a custom idempotency key for this request
        """
        options = make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                idempotency_key=idempotency_key,
        )
        if max_retries is not NOT_GIVEN:
            options["max_retries"] = max_retries

        return await self._post(
            "/api/v1/sample",
            body=await async_maybe_transform(
                {
                    "num_samples": num_samples,
                    "prompt": prompt,
                    "sampling_params": sampling_params,
                    "base_model": base_model,
                    "model_path": model_path,
                    "prompt_logprobs": prompt_logprobs,
                    "type": type,
                },
                sampling_sample_params.SamplingSampleParams,
            ),
            options=options,
            cast_to=SampleResponse,
        )


class SamplingResourceWithRawResponse:
    def __init__(self, sampling: SamplingResource) -> None:
        self._sampling = sampling

        self.asample = to_raw_response_wrapper(
            sampling.asample,
        )
        self.sample = to_raw_response_wrapper(
            sampling.sample,
        )


class AsyncSamplingResourceWithRawResponse:
    def __init__(self, sampling: AsyncSamplingResource) -> None:
        self._sampling = sampling

        self.asample = async_to_raw_response_wrapper(
            sampling.asample,
        )
        self.sample = async_to_raw_response_wrapper(
            sampling.sample,
        )


class SamplingResourceWithStreamingResponse:
    def __init__(self, sampling: SamplingResource) -> None:
        self._sampling = sampling

        self.asample = to_streamed_response_wrapper(
            sampling.asample,
        )
        self.sample = to_streamed_response_wrapper(
            sampling.sample,
        )


class AsyncSamplingResourceWithStreamingResponse:
    def __init__(self, sampling: AsyncSamplingResource) -> None:
        self._sampling = sampling

        self.asample = async_to_streamed_response_wrapper(
            sampling.asample,
        )
        self.sample = async_to_streamed_response_wrapper(
            sampling.sample,
        )



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

from enum import Enum


class ClientConnectionPoolType(Enum):
    SAMPLE = "sample"
    TRAIN = "train"
    RETRIEVE_PROMISE = "retrieve_promise"
    TELEMETRY = "telemetry"



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

import logging
from typing import Any, Dict, List, Sequence, Set, cast

import numpy as np
from hpcai.types import ForwardBackwardOutput, LossFnOutput

logger = logging.getLogger(__name__)

Metrics = Dict[str, float]


def combine_fwd_bwd_output_results(
    results: Sequence[ForwardBackwardOutput],
) -> ForwardBackwardOutput:
    if not results:
        return ForwardBackwardOutput(loss_fn_output_type="", metrics={}, loss_fn_outputs=[])

    combined_metrics = _metrics_reduction(results)
    combined_outputs = _combine_loss_fn_outputs(results)

    return ForwardBackwardOutput(
        loss_fn_output_type=results[0].loss_fn_output_type,
        metrics=combined_metrics,
        loss_fn_outputs=combined_outputs,
    )


def _combine_loss_fn_outputs(results: Sequence[ForwardBackwardOutput]) -> List[LossFnOutput]:
    return [output for result in results for output in result.loss_fn_outputs]


def _order_insensitive_hash(xs: Sequence[Set[Any]] | Sequence[float]) -> int:
    """Combine hash values in an order-insensitive way.

    Args:
        xs: Either a sequence of sets (original data) or a sequence of already-computed hash values
    """
    # If we have sets, flatten and hash them (original behavior)
    if xs and isinstance(xs[0], set):
        return hash(tuple(sorted([y for x in xs for y in cast(Set[Any], x)])))

    # If we have already-computed hash values, combine them deterministically
    # Sort them to ensure order-insensitive combination
    return hash(tuple(sorted(int(cast(float, x)) for x in xs)))


def _mean(xs: Sequence[float | int], weights: Sequence[float] | None = None) -> float:
    if weights is None:
        return np.mean(xs).item()
    return np.average(xs, weights=weights).item()


def _sum(xs: Sequence[float | int]) -> float:
    return np.sum(xs).item()


def _min(xs: Sequence[float | int]) -> float:
    return np.min(xs).item()


def _max(xs: Sequence[float | int]) -> float:
    return np.max(xs).item()


def _slack(xs: Sequence[float | int], weights: Sequence[float] | None = None) -> float:
    if weights is None:
        return (np.max(xs) - np.mean(xs)).item()
    return (np.max(xs) - np.average(xs, weights=weights)).item()


def _unique(xs: Sequence[float | int]) -> Sequence[float | int]:
    """
    A unique metric can't actually fold. But in order to work around the fact that
    it's a str:float dict, we just insert unique keys with some suffix for each
    unique value. It's a hack, that's for sure.

    This is a dummy identity function just for documentation and consistency purposes.
    """
    return xs


REDUCE_MAP = {
    "mean": _mean,
    "sum": _sum,
    "min": _min,
    "max": _max,
    "slack": _slack,
    "hash_unordered": _order_insensitive_hash,
    "unique": _unique,
}


def _metrics_reduction(results: Sequence[ForwardBackwardOutput]) -> Metrics:
    """Reduce metrics from all actors.
    every metric must indicate a reduction_type in its name for example "mfu:mean"

    Metrics are weighted by the number of loss_fn_outputs (data points) each actor processed.
    """
    if not results:
        return {}
    keys = results[0].metrics.keys()

    weights = [len(m.loss_fn_outputs) for m in results]

    res = {}
    for key in keys:
        name, reduction = key.split(":")
        if reduction not in REDUCE_MAP:
            # Can happen when a new reduction type is added
            logger.debug(
                f"Invalid {reduction=} for metric {name=}. Expecting one of {REDUCE_MAP.keys()}"
            )
            continue
        if not all(key in m.metrics for m in results):
            continue
        reduce_fn = REDUCE_MAP[reduction]
        values = [m.metrics[key] for m in results]

        if reduction in ["mean", "slack"]:
            res[key] = reduce_fn(values, weights)
        elif reduction in ["unique"]:
            res[key] = values[0]
            res.update({f"{key}_{i + 1}": v for i, v in enumerate(values[1:])})
        else:
            res[key] = reduce_fn(values)
    return res



# Copyright 2025 Thinking Machines Lab
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

from __future__ import annotations

import asyncio
from collections.abc import Coroutine
from contextlib import AbstractContextManager
from typing import Any, Protocol, TypeVar

from hpcai._client import AsyncHpcAI

from .public_interfaces.api_future import AwaitableConcurrentFuture
from .client_connection_pool_type import ClientConnectionPoolType

T = TypeVar("T")


class AsyncHpcAIProvider(Protocol):
    # both of the following methods should be threadsafe
    def get_loop(self) -> asyncio.AbstractEventLoop: ...

    def run_coroutine_threadsafe(
        self,
        coro: Coroutine[Any, Any, T],
    ) -> AwaitableConcurrentFuture[T]: ...

    # must be called and used within the provided event loop
    def aclient(
        self, client_pool_type: ClientConnectionPoolType
    ) -> AbstractContextManager[AsyncHpcAI]: ...



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from .._compat import PYDANTIC_V2, ConfigDict
from .._models import StrictBase
from .model_input import ModelInput
from .sampling_params import SamplingParams

__all__ = ["SampleRequest"]


class SampleRequest(StrictBase):
    num_samples: int = 1
    """Number of samples to generate"""

    prompt: ModelInput

    sampling_params: SamplingParams

    base_model: Optional[str] = None
    """Optional base model name to sample from.

    Is inferred from model_path, if provided. If sampling against a base model, this
    is required.
    """

    model_path: Optional[str] = None
    """Optional hpcai:// path to your model weights or LoRA weights.

    If not provided, samples against the base model.
    """

    prompt_logprobs: Optional[bool] = None
    """If set to `true`, computes and returns logprobs on the prompt tokens.

    Defaults to false.
    """

    type: Optional[Literal["sample"]] = None

    if PYDANTIC_V2:
        # allow fields with a `model_` prefix
        model_config = ConfigDict(protected_namespaces=tuple())



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

from .._models import BaseModel
from .checkpoint import Checkpoint

__all__ = ["CheckpointsListResponse"]


class CheckpointsListResponse(BaseModel):
    checkpoints: list[Checkpoint]
    """List of available model checkpoints for the model"""



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal, Required, TypedDict

from .model_input_param import ModelInputParam
from .sampling_params_param import SamplingParamsParam

__all__ = ["SamplingSampleParams"]


class SamplingSampleParams(TypedDict, total=False):
    num_samples: int
    """Number of samples to generate"""

    prompt: Required[ModelInputParam]

    sampling_params: Required[SamplingParamsParam]

    base_model: str
    """Optional base model name to sample from.

    Is inferred from model_path, if provided. If sampling against a base model, this
    is required.
    """

    model_path: str
    """Optional hpcai:// path to your model weights or LoRA weights.

    If not provided, samples against the base model.
    """

    prompt_logprobs: bool
    """If set to `true`, computes and returns logprobs on the prompt tokens.

    Defaults to false.
    """

    type: Literal["sample"]



# Copyright 2025 Thinking Machines Lab
#
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

from __future__ import annotations

import sys
import asyncio
import functools
import contextvars
from typing import Any, TypeVar, Callable, Awaitable
from typing_extensions import ParamSpec

import anyio
import sniffio
import anyio.to_thread

T_Retval = TypeVar("T_Retval")
T_ParamSpec = ParamSpec("T_ParamSpec")


if sys.version_info >= (3, 9):
    _asyncio_to_thread = asyncio.to_thread
else:
    # backport of https://docs.python.org/3/library/asyncio-task.html#asyncio.to_thread
    # for Python 3.8 support
    async def _asyncio_to_thread(
        func: Callable[T_ParamSpec, T_Retval], /, *args: T_ParamSpec.args, **kwargs: T_ParamSpec.kwargs
    ) -> Any:
        """Asynchronously run function *func* in a separate thread.

        Any *args and **kwargs supplied for this function are directly passed
        to *func*. Also, the current :class:`contextvars.Context` is propagated,
        allowing context variables from the main thread to be accessed in the
        separate thread.

        Returns a coroutine that can be awaited to get the eventual result of *func*.
        """
        loop = asyncio.events.get_running_loop()
        ctx = contextvars.copy_context()
        func_call = functools.partial(ctx.run, func, *args, **kwargs)
        return await loop.run_in_executor(None, func_call)


async def to_thread(
    func: Callable[T_ParamSpec, T_Retval], /, *args: T_ParamSpec.args, **kwargs: T_ParamSpec.kwargs
) -> T_Retval:
    if sniffio.current_async_library() == "asyncio":
        return await _asyncio_to_thread(func, *args, **kwargs)

    return await anyio.to_thread.run_sync(
        functools.partial(func, *args, **kwargs),
    )


# inspired by `asyncer`, https://github.com/tiangolo/asyncer
def asyncify(function: Callable[T_ParamSpec, T_Retval]) -> Callable[T_ParamSpec, Awaitable[T_Retval]]:
    """
    Take a blocking function and create an async one that receives the same
    positional and keyword arguments. For python version 3.9 and above, it uses
    asyncio.to_thread to run the function in a separate thread. For python version
    3.8, it uses locally defined copy of the asyncio.to_thread function which was
    introduced in python 3.9.

    Usage:

    ```python
    def blocking_func(arg1, arg2, kwarg1=None):
        # blocking code
        return result


    result = asyncify(blocking_function)(arg1, arg2, kwarg1=value1)
    ```

    ## Arguments

    `function`: a blocking regular callable (e.g. a function)

    ## Return

    An async function that takes the same positional and keyword arguments as the
    original one, that when called runs the same original function in a thread worker
    and returns the result.
    """

    async def wrapper(*args: T_ParamSpec.args, **kwargs: T_ParamSpec.kwargs) -> T_Retval:
        return await to_thread(function, *args, **kwargs)

    return wrapper



# Copyright 2025 Thinking Machines Lab
# Licensed under the Apache License, Version 2.0
#
# Modifications:
# - Adapted for HPC-AI cloud fine-tuning workflow
# Copyright © 2025 HPC-AI.COM

import torch
from hpcai.types import Datum, ModelInput, TensorData
from hpcai.cookbook.renderers import Message, Renderer, TrainOnWhat
from hpcai.cookbook.common import datum_from_tokens_weights


def datum_from_tokens_weights(
    tokens: torch.Tensor,
    weights: torch.Tensor,
    max_length: int | None = None,
) -> Datum:
    if max_length is not None:
        tokens = tokens[:max_length]
    weights = weights[:max_length]

    input_tokens = tokens[:-1]
    target_tokens = tokens[1:]
    weights = weights[1:]

    return Datum(
        model_input=ModelInput.from_ints(tokens=input_tokens.tolist()),
        loss_fn_inputs={
            "weights": TensorData(
                data=weights.tolist(),
                dtype="float32",
                shape=list(weights.shape),
            ),
            "target_tokens": TensorData(
                data=[int(x) for x in target_tokens.tolist()],
                dtype="int64",
                shape=list(target_tokens.shape),
            ),
        },
    )


def conversation_to_datum(
    conversation: list[Message],
    renderer: Renderer,
    max_length: int | None,
    train_on_what: TrainOnWhat = TrainOnWhat.ALL_ASSISTANT_MESSAGES,
) -> Datum:
    """Common function to process a list of messages into a Datum."""
    tokens, weights = renderer.build_supervised_example(conversation, train_on_what=train_on_what)
    return datum_from_tokens_weights(tokens, weights, max_length)
